//  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "nelder_mead.h"
#include "line_sampling_ported.h"
#include <iostream>
#include <algorithm>
#include <fstream>
#include <random>
#include <cstdint>
#include <thread>
#include <mutex>
#include <utility>
#include <chrono>
#include <memory>
#include <limits>


//! Describes a single test case for the sampling procedure, i.e. a
//! configuration in which one sample can be produced
struct test_case_t {
	//! The parameter values from which this test case has been constructed
	search_vec_t parameters;
	//! The line to be sampled
	projected_solid_angle_line_t line;
	//! The normalized line direction vector
	vec3 line_direction;
	//! The random number number used to sample the line
	Float random_number;
	//! The sampled direction
	vec3 sampled_dir;
};


//! Produces a sample for the given test case
vec3 sample_test_case(test_case_t test) {
	return sample_projected_solid_angle_line(test.line, test.random_number);
}


//! Turns a vector of four numbers between zero and one into a test case for
//! the sampling procedure
inline test_case_t construct_test_case(search_vec_t input, bool compute_sampled_dir=false) {
	test_case_t test;
	test.parameters = input;
	// Clamp the input
	for (uint i = 0; i != DIMENSIONALITY; ++i)
		input[i] = clamp(input[i], 0.0, 1.0);
	// The shading normal is in the xy-plane
	vec3 shading_normal = make_vec3(cos(input[0] * M_PI), sin(input[0] * M_PI), 0.0);
	// The endpoints of the linear light are along the line with y=1, z=0 and
	// above the horizon
	Float intersection_x = -shading_normal.y / shading_normal.x;
	Float min_x = (shading_normal.x > 0.0) ? intersection_x : std::numeric_limits<Float>::min();
	Float max_x = (shading_normal.x < 0.0) ? intersection_x : std::numeric_limits<Float>::max();
	Float min_angle = std::atan(min_x);
	Float max_angle = std::atan(max_x);
	// We limit lines to an opening angle of 89.99999 degrees from the closest
	// point
	const Float bound_radians = 1.5707961522619714197983639992706745932;
	min_angle = clamp(min_angle, -bound_radians, bound_radians);
	max_angle = clamp(max_angle, -bound_radians, bound_radians);
	Float angle_begin = min_angle * (1.0 - input[1]) + max_angle * input[1];
	Float angle_end = angle_begin * (1.0 - input[2]) + max_angle * input[2];
	vec3 line_begin = make_vec3(std::tan(angle_begin), 1.0, 0.0);
	vec3 line_end = make_vec3(std::tan(angle_end), 1.0, 0.0);
	test.line_direction = make_vec3(1.0, 0.0, 0.0);
	// Prepare the line for sampling
	test.line = prepare_projected_solid_angle_line_sampling(shading_normal, line_begin, test.line_direction, line_end.x - line_begin.x);
	// Finally, we need a random number
	test.random_number = input[3];
	// Optionally take a sample
	if (compute_sampled_dir)
		test.sampled_dir = sample_test_case(test);
	return test;
}


//! Holds the worst error encountered by a thread and meta-data about the
//! effort invested to find it
struct thread_progress_t {
public:
	//! Initializes with zero error and no effort
	thread_progress_t() :
	worst_error(0.0),
	invocation_count(0),
	init_count(0) {
	}

	//! Adds up effort and takes the worst error between this object and the
	//! given one
	void merge(const thread_progress_t& rhs) {
		if (worst_error < rhs.worst_error) {
			worst_error = rhs.worst_error;
			worst_parameters = rhs.worst_parameters;
		}
		invocation_count += rhs.invocation_count;
		init_count += rhs.init_count;
	}

	//! The worst error encountered by this thread thus far
	Float worst_error;
	//! The parameters that lead to the worst error encountered thus far
	search_vec_t worst_parameters;
	//! The number of objective function invocations that this thread has
	//! performed
	uint64_t invocation_count;
	//! The number of random initializations that this thread has used
	uint64_t init_count;
};


//! A thread that will search worst case errors indefinitely
class thread_t : public thread {
public:
	//! Launches the thread
	thread_t(random_device& random_device) :
	random_generator(random_device()),
	uniform_distribution(0.0, 1.0) {
		// Launch the thread
		thread::operator=(thread(
			[this]() {
				while (true)
					try_random_initialization();
			}));
	}

	//! Generates and tries a single random initialization and updates the
	//! progress
	void try_random_initialization() {
		// Generate a random initialization
		search_vec_t initialization;
		for (uint i = 0; i != DIMENSIONALITY; ++i)
			initialization[i] = uniform_distribution(random_generator);
		// Use Nelder-Mead to maximize the error
		uint64_t invocation_count = 0;
		auto objective = [this, &invocation_count](const search_vec_t& point) -> Float {
			++invocation_count;
			test_case_t test = construct_test_case(point);
			// With extremely short lines, we run into numerical issues but
			// this experiment is concerned with theoretical convergence
			// properties. These lines are so short that single-precision
			// floats could barely resolve the difference between begin and
			// end.
			if (test.line.solid.extent_s < 1.0e-7)
				return 0.0;
			// Take a sample
			vec3 sample = sample_test_case(test);
			// Compute the error of the result
			Float result = compute_projected_solid_angle_line_sampling_error(test.line, test.random_number, sample);
			return -abs(result);
		};
		Float error;
		search_vec_t optimized = minimize_nelder_mead(&error, initialization, objective);
		error = -error;
		// Update the progress
		progress_mutex.lock();
		progress.invocation_count += invocation_count;
		++progress.init_count;
		if (error > progress.worst_error) {
			progress.worst_error = error;
			progress.worst_parameters = optimized;
		}
		progress_mutex.unlock();
	}

	//! A random generator with a seed specific to this thread
	mt19937_64 random_generator;
	//! A uniform distribution on [0,1)
	uniform_real_distribution<Float> uniform_distribution;
	//! A mutex that should be acquired when reading or writing attributes of
	//! progress
	mutex progress_mutex;
	//! The progress of this thread
	thread_progress_t progress;
};


int main() {
	// Warn about Visual C++
	if (sizeof(long double) < 10) {
		cout << "WARNING: This program needs to be compiled in an environment where long double" << endl;
		cout << "uses at least 63 bits for the mantissa. Do not use Visual C++." << endl;
	}
	// Prepare truly random bits
	random_device random_device;

	// Take a sample with the worst case found previously to enable analysis
	if (true) {
		ifstream test_case_file("test_case.dat", ios_base::binary);
		if (test_case_file.is_open()) {
			test_case_t test;
			test_case_file.read(reinterpret_cast<char*>(&test), sizeof(test));
			test = construct_test_case(test.parameters);
			vec3 sample = sample_test_case(test);
			Float error = abs(compute_projected_solid_angle_line_sampling_error(test.line, test.random_number, sample));
			cout << "Error in a previous run: " << error << endl;
		}
	}
	// Compute the median error
	if (false) {
		// Prepare a random generator
		mt19937_64 random_generator(random_device());
		uniform_real_distribution<Float> uniform_distribution(0.0f, 1.0f);
		// Generate samples with random test cases and record their errors
		uint iteration_count = 2;
		size_t sample_count = 100000000;
		vector<Float> errors(sample_count);
		std::cout << "Using " << iteration_count << " iterations and " << sample_count << " samples to compute percentiles of errors." << std::endl;
		for (size_t i = 0; i != sample_count; ++i) {
			search_vec_t random_numbers;
			for (uint j = 0; j != DIMENSIONALITY; ++j)
				random_numbers[j] = uniform_distribution(random_generator);
			test_case_t test = construct_test_case(random_numbers);
			vec3 sample = sample_test_case(test);
			errors[i] = abs(compute_projected_solid_angle_line_sampling_error(test.line, test.random_number, sample));
		}
		// Sort and report percentiles
		uint percentiles_times_ten[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50, 100, 200, 300, 400, 500, 600, 700, 800, 900, 950, 960, 970, 980, 990, 995, 996, 997, 998, 999};
		std::cout << "Sorting the errors." << std::endl;
		sort(errors.begin(), errors.end());
		for (uint percentile_times_ten : percentiles_times_ten) {
			size_t index = (percentile_times_ten * sample_count) / 1000;
			std::cout << percentile_times_ten * 0.1 << "th percentile of backward error: " << errors[index] << std::endl;
		}
	}

	// Run as many threads as the hardware can handle
	vector<unique_ptr<thread_t> > threads;
	uint thread_count = thread::hardware_concurrency();
	for (uint i = 0; i != thread_count; ++i)
		threads.emplace_back(new thread_t(random_device));
	// Report on their progress in regular intervals
	Float worst_error = 0.0;
	uint64_t reported_invocation_count = 0;
	while (true) {
		// Combine the progress of all threads
		thread_progress_t total_progress;
		for (const unique_ptr<thread_t>& thread : threads) {
			thread->progress_mutex.lock();
			total_progress.merge(thread->progress);
			thread->progress_mutex.unlock();
		}
		// Report the error if it is worse than anything seen thus far
		if (total_progress.worst_error > worst_error) {
			worst_error = total_progress.worst_error;
			cout << "Error: " << worst_error << endl;
			cout << "Test case:";
			for (uint i = 0; i != DIMENSIONALITY; ++i)
				cout << " " << total_progress.worst_parameters[i];
			cout << endl;
			cout << total_progress.invocation_count / 1000000 << " million invocations and " << total_progress.init_count << " initializations." << endl;
			test_case_t test = construct_test_case(total_progress.worst_parameters, true);
			ofstream("test_case.dat", ios_base::binary).write(reinterpret_cast<char*>(&test), sizeof(test));
		}
		// Report the invocation count and the initialization count regularly
		if (total_progress.invocation_count / 5000000000 > reported_invocation_count) {
			cout << total_progress.invocation_count / 1000000000 << " billion invocations and " << total_progress.init_count << " initializations." << endl;
			++reported_invocation_count;
		}
		// This thread should not stress the CPU
		this_thread::sleep_for(chrono::seconds());
	}
}
