//  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.


#pragma once
//! A PBRT-style typedef to stay flexible with regard to precision. We expect
//! long double to have a 63-bit mantissa (80 bits total).
typedef long double Float;
#include "mini_glm.h"
#include <cmath>

using namespace std;

#define GLSL_MATH_FUNCTION inline

typedef unsigned int uint;

GLSL_MATH_FUNCTION Float inversesqrt(Float value) {
	return 1.0 / sqrt(value);
}

GLSL_MATH_FUNCTION Float min(Float lhs, Float rhs) {
	return (lhs < rhs) ? lhs : rhs;
}

GLSL_MATH_FUNCTION Float max(Float lhs, Float rhs) {
	return (lhs > rhs) ? lhs : rhs;
}

GLSL_MATH_FUNCTION Float dot(vec2 lhs, vec2 rhs) {
	return lhs.x * rhs.x + lhs.y * rhs.y;
}

GLSL_MATH_FUNCTION Float dot(vec3 lhs, vec3 rhs) {
	return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z;
}


GLSL_MATH_FUNCTION vec3 cross(vec3 lhs, vec3 rhs) {
	vec3 result = {
		lhs.y * rhs.z - lhs.z * rhs.y,
		lhs.z * rhs.x - lhs.x * rhs.z,
		lhs.x * rhs.y - lhs.y * rhs.x,
	};
	return result;
}

GLSL_MATH_FUNCTION vec2 normalize(vec2 rhs) {
	return rhs * inversesqrt(dot(rhs, rhs));
}

GLSL_MATH_FUNCTION vec3 normalize(vec3 rhs) {
	return rhs * inversesqrt(dot(rhs, rhs));
}

GLSL_MATH_FUNCTION Float length(vec3 rhs) {
	return sqrt(dot(rhs, rhs));
}


GLSL_MATH_FUNCTION Float determinant(mat2 matrix) {
	return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0];
}


GLSL_MATH_FUNCTION mat2 outerProduct(vec2 lhs, vec2 rhs) {
	return mat2(
		lhs[0] * rhs[0], lhs[1] * rhs[0], 
		lhs[0] * rhs[1], lhs[1] * rhs[1]
	);
}


GLSL_MATH_FUNCTION Float atan(Float y, Float x) {
	return atan2(y, x);
}

GLSL_MATH_FUNCTION vec2 fma(vec2 factor_0, vec2 factor_1, vec2 summand) {
	return make_vec2(
		fma(factor_0.x, factor_1.x, summand.x),
		fma(factor_0.y, factor_1.y, summand.y));
}

GLSL_MATH_FUNCTION vec3 fma(vec3 factor_0, vec3 factor_1, vec3 summand) {
	return make_vec3(
		fma(factor_0.x, factor_1.x, summand.x),
		fma(factor_0.y, factor_1.y, summand.y),
		fma(factor_0.z, factor_1.z, summand.z));
}

GLSL_MATH_FUNCTION Float clamp(Float value, Float min, Float max) {
	value = (value < min) ? min : value;
	value = (value > max) ? max : value;
	return value;
}
