//  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "glsl_math.h"

#include "math_constants.glsl"


/*! This structure carries intermediate results that only need to be computed
	once per line and shading point to take samples proportional to solid
	angle.*/
struct solid_angle_line_t {
	//! The vector from the shading point to the beginning of the line
	vec3 begin;
	//! The normalized vector from begin to end of the line
	vec3 line_direction;
	//! The reciprocal of the minimal distance between the shading point and
	//! the infinitely extended line
	Float rcp_min_distance;
	//! dot(normalize(line_direction), begin)
	Float scaled_begin_s;
	//! dot(normalize(line_direction), normalize(begin))
	Float begin_s;
	//! dot(normalize(line_direction), normalize(end) - normalize(begin))
	Float extent_s;
	/*! The limit of the solid angle of the cyliner, divided by its radius or
		the limit of the projected solid angle of the cyliner, divided by its
		radius if this struct is used for projected solid angle sampling.*/
	Float rcp_density;
};


/*! This structure carries intermediate results that only need to be computed
	once per line and shading point to take samples proportional to projected
	solid angle.*/
struct projected_solid_angle_line_t {
	//! We reuse functionality implemented for solid angle sampling
	solid_angle_line_t solid;
	/*! The first component is dot(line_direction, surface_normal), the second
		is dot(closest_point, surface_normal) where closest point is a
		normalized vector from the shading point to the closest point on the
		infinite line.*/
	vec2 normal;
	//! The indefinite CDF at begin_s (see evaluate_line_sampling_cdf())
	Float begin_cdf;
	//! The CDF at begin_s + extent_s minus begin_cdf
	Float extent_cdf;
};


/*! Clips the specified line against the plane through the given origin with
	the given normal. The positive half space is kept. All input parameters
	are updated to describe the clipped line after return. line_direction must
	be normalized. If the given line is completely below the horizon,
	line_length becomes exactly zero.*/
inline void clip_line(vec3 plane_origin, vec3 plane_normal, vec3& line_begin, vec3& line_end, vec3 line_direction, Float& line_length) {
	Float normal_s = dot(plane_normal, line_direction);
	Float normal_dot_line_begin = dot(plane_normal, line_begin - plane_origin);
	Float intersection_s = -normal_dot_line_begin / normal_s;
	Float begin_s = (normal_dot_line_begin < 0.0f) ? intersection_s : 0.0f;
	Float end_s = (fma(line_length, normal_s, normal_dot_line_begin) < 0.0f) ? intersection_s : line_length;
	line_end = fma(line_direction, make_vec3(end_s), line_begin);
	line_begin = fma(line_direction, make_vec3(begin_s), line_begin);
	line_length = end_s - begin_s;
}


/*!	Prepares all intermediate values to sample a cylinder of infinitesimally
	small radius proportional to solid angle.
	\see prepare_projected_solid_angle_line_sampling() */
inline solid_angle_line_t prepare_solid_angle_line_sampling(vec3 line_begin, vec3 line_direction, Float line_length) {
	solid_angle_line_t line;
	// Copy inputs
	line.begin = line_begin;
	line.line_direction = line_direction;
	// Compute the minimal distance to the line
	line.scaled_begin_s = dot(line_direction, line_begin);
	Float sq_begin_distance = dot(line_begin, line_begin);
	Float sq_min_distance = fma(-line.scaled_begin_s, line.scaled_begin_s, sq_begin_distance);
	line.rcp_min_distance = inversesqrt(sq_min_distance);
	// Compute the extent along the line direction of normalized direction
	// vectors to line vertices
	line.begin_s = line.scaled_begin_s * inversesqrt(sq_begin_distance);
	Float scaled_end_s = line.scaled_begin_s + line_length;
	Float rcp_end_distance = inversesqrt(fma(scaled_end_s, scaled_end_s, sq_min_distance));
	line.extent_s = fma(scaled_end_s, rcp_end_distance, -line.begin_s);
	// Compute the solid angle over radius
	line.rcp_density = 2.0f * line.rcp_min_distance * line.extent_s;
	return line;
}


/*! Given the output of prepare_solid_angle_line_sampling(), this function maps
	the given random number in the range from 0 to 1 to a normalized direction
	vector providing a sample of the line in the original space (used for
	arguments of prepare_solid_angle_line_sampling()). Samples are distributed
	in proportion to solid angle assuming uniform inputs.*/
inline vec3 sample_solid_angle_line(solid_angle_line_t line, Float random_number) {
	Float s = fma(random_number, line.extent_s, line.begin_s);
	Float t = sqrt(max(0.0f, fma(-s, s, 1.0f)));
	Float y = line.rcp_min_distance * t;
	Float x = fma(-y, line.scaled_begin_s, s);
	return x * line.line_direction + y * line.begin;
}


/*! Evaluates the unnormalized CDF for projected solid angle sampling of lines.
	\param normal The s- and t-coordinate of the normalized shading normal.
	\param s The s-coordinate where the CDF is evaluated.
	\param cdf_begin This value gets subtracted from the result. Pass 0.0f if
		you want an indefinite CDF. Pass the indefinite CDF for the beginning
		of the line, if you want a definite (but unnormalized) CDF. The
		subraction is free thanks to fused-multiply add.
	\return The unnormalized CDF, i.e.:
		normal.x * s*s + normal.y * (sqrt(1-s*s) * s + asin(s)) - cdf_begin */
inline Float evaluate_line_sampling_cdf(vec2 normal, Float s, Float cdf_begin) {
	s = clamp(s, -1.0f, 1.0f);
	Float angle = asin(s);
	Float t = sqrt(fma(-s, s, 1.0f));
	return fma(dot(normal, make_vec2(s, t)), s, fma(angle, normal.y, -cdf_begin));
}


//! Same as prepare_projected_solid_angle_line_sampling() but assumes that the
//! given line is completely above the horizon, e.g. because it has been
//! clipped already. It also assumes that the line length is greater zero.
inline projected_solid_angle_line_t prepare_projected_solid_angle_line_sampling_without_clipping(vec3 surface_normal, vec3 line_begin, vec3 line_direction, Float line_length) {
	projected_solid_angle_line_t line;
	// We reuse methods for solid angle sampling
	line.solid = prepare_solid_angle_line_sampling(line_begin, line_direction, line_length);
	// Evaluate the projected solid angle integral at the relevant values of s
	line.normal.x = dot(surface_normal, line_direction);
	line.normal.y = fma(-line.solid.scaled_begin_s, line.normal.x, dot(surface_normal, line_begin)) * line.solid.rcp_min_distance;
	// Determine the CDF for random number 0 and 1
	line.begin_cdf = evaluate_line_sampling_cdf(line.normal, line.solid.begin_s, 0.0f);
	line.extent_cdf = evaluate_line_sampling_cdf(line.normal, line.solid.begin_s + line.solid.extent_s, line.begin_cdf);
	// Overwrite solid angle with the the projected solid angle (per radius)
	line.solid.rcp_density = line.solid.rcp_min_distance * abs(line.extent_cdf);
	// It is possible that the minimal distance is 0 (along the line) and then
	// rcp_density is NaN but should be 0
	line.solid.rcp_density = (line.solid.rcp_density > 0.0f) ? line.solid.rcp_density : 0.0f;
	return line;
}


/*!	Prepares all intermediate values to sample a cylinder of infinitesimally
	small radius proportional to projected solid angle. The line begins at
	line_begin and ends at line_begin + line_length * line_direction. The
	shading point is in the origin, direction vectors are normalized.*/
inline projected_solid_angle_line_t prepare_projected_solid_angle_line_sampling(vec3 surface_normal, vec3 line_begin, vec3 line_direction, Float line_length) {
	projected_solid_angle_line_t line;
	// This will hopefully be optimized away
	line.solid.begin = line_begin;
	line.solid.line_direction = line_direction;
	// Clip the given line
	vec3 line_end;
	clip_line(make_vec3(0.0f), surface_normal, line_begin, line_end, line_direction, line_length);
	if (line_length <= 0.0f) {
		line.solid.rcp_density = 0.0f;
		// GLSL permits uninitialized variables, HLSL does not
		// line.solid.rcp_min_distance = line.solid.scaled_begin_s = line.solid.begin_s = line.solid.extent_s = 0.0f;
		// line.normal = make_vec2(0.0f);
		// line.begin_cdf = line.extent_cdf = 0.0f;
		return line;
	}
	// Forward to the variant without clipping
	else {
		return prepare_projected_solid_angle_line_sampling_without_clipping(surface_normal, line_begin, line_direction, line_length);
	}
}


//! \return true iff the given line segment is completely below the horizon
inline bool is_projected_solid_angle_line_empty(projected_solid_angle_line_t line) {
	return line.solid.rcp_density <= 0.0f;
}


/*! Returns a solution to the given homogeneous quadratic equation, i.e. a
	non-zero vector root such that dot(root, quadratic * root) == 0.0f. The
	returned root depends continuously on quadratic. Pass -quadratic if you
	want the other root. quadratic must be symmetric.
	\note The implementation is as proposed by Blinn, except that we do not
	have a special case for quadratic[0][1] + quadratic[1][0] == 0.0f. Unlike
	the standard quadratic formula, it allows us to postpone a division and is
	stable in all cases.
	James F. Blinn 2006, How to Solve a Quadratic Equation, Part 2, IEEE
	Computer Graphics and Applications 26:2 https://doi.org/10.1109/MCG.2006.35
*/
inline vec2 solve_symmetric_homogeneous_quadratic(mat2 quadratic) {
	Float coeff_xy = quadratic[0][1];
	Float sqrt_discriminant = sqrt(max(0.0f, coeff_xy * coeff_xy - quadratic[0][0] * quadratic[1][1]));
	Float scaled_root = abs(coeff_xy) + sqrt_discriminant;
	return (coeff_xy >= 0.0f) ? make_vec2(scaled_root, -quadratic[0][0]) : make_vec2(quadratic[1][1], scaled_root);
}


/*! Does the bulk of the work for projected solid angle sampling, namely
	inverting the relevant CDF for inverse function sampling. You also have to
	pass the random number between 0 and 1 that was used to compute cdf.
	\return The s- and t-coordinate for the normalized direction.*/
inline vec2 invert_line_sampling_cdf(projected_solid_angle_line_t line, Float random_number, Float cdf) {
	// Use solid angle sampling for the initialization
	vec2 initial;
	initial.x = fma(random_number, line.solid.extent_s, line.solid.begin_s);
	initial.x = clamp(initial.x, -1.0f, 1.0f);
	initial.y = sqrt(fma(-initial.x, initial.x, 1.0f));
	vec2 dir = initial;
	// Refine iteratively. Two iterations are known to give sufficiently good
	// results everywhere. If you want to be really sure, increase to three,
	// which will give you more precision than single-precision floats can
	// handle almost everywhere.
	Float angle = asin(dir.x);
	#pragma unroll
	for (uint i = 0; i != 2; ++i) {
		// Evaluate the objective function and two derivatives
		Float dot_0 = dot(line.normal, dir);
		Float objective = fma(dot_0, dir.x, fma(angle, line.normal.y, -cdf));
		Float half_derivative_1 = dot_0 * dir.y;
		Float dot_1 = dot(line.normal, make_vec2(dir.y, -dir.x));
		Float half_derivative_2 = dot_1 * dir.y - dot_0 * dir.x;
		// Compute the root of minimal magnitude for a quadratic Taylor
		// expansion. This is Laguerre's method for n = 2, also known as
		// Halley's irrational formula.
		mat2 quadratic = mat2(half_derivative_2, half_derivative_1, half_derivative_1, objective);
		vec2 root = solve_symmetric_homogeneous_quadratic(-quadratic);
		// On to the next iteration, or maybe we are done
		angle += root.x / root.y;
		angle = clamp(angle, -M_HALF_PI, M_HALF_PI);
		dir.x = sin(angle);
		dir.y = cos(angle);
	}
	// In boundary cases the iteration is unstable but not needed
	Float acceptable_error = 1.0e-5f;
	bool use_iterations = (abs(random_number - 0.5f) <= 0.5f - acceptable_error);
	dir.x = use_iterations ? dir.x : initial.x;
	dir.y = use_iterations ? dir.y : initial.y;
	return dir;
}


/*! Given the output of prepare_projected_solid_angle_line_sampling(), this
	function maps the given random number in the range from 0 to 1 to a
	normalized direction vector providing a sample of the line in the original
	space (used for arguments of
	prepare_projected_solid_angle_line_sampling()). Samples are distributed in
	proportion to projected solid angle assuming uniform inputs.*/
inline vec3 sample_projected_solid_angle_line(projected_solid_angle_line_t line, Float random_number) {
	// Figure out what the CDF is supposed to be
	Float cdf = fma(random_number, line.extent_cdf, line.begin_cdf);
	// Invert
	vec2 dir = invert_line_sampling_cdf(line, random_number, cdf);
	// Turn it into a sample on the great circle of the light
	Float y = line.solid.rcp_min_distance * dir.y;
	Float x = fma(-y, line.solid.scaled_begin_s, dir.x);
	return x * line.solid.line_direction + y * line.solid.begin;
}


/*! Determines the error of a sample from the projected solid angle of an
	infinitesimally thin cylinder due to the iterative procedure.
	\param line Output of prepare_projected_solid_angle_line_sampling().
	\param random_numbers The value passed for random_number in
		sample_projected_solid_angle_line().
	\param sampled_dir The direction returned by
		sample_projected_solid_angle_line().
	\return The signed backward error, i.e. the difference between
		random_number and the random number that would yield sampled_dir with
		perfect computation. Note that the result is subject to rounding error
		itself.*/
inline Float compute_projected_solid_angle_line_sampling_error(projected_solid_angle_line_t line, Float random_number, vec3 sampled_dir) {
	// Figure out what the CDF is supposed to be
	Float cdf = fma(random_number, line.extent_cdf, line.begin_cdf);
	// Compute the s-coordinate of the sampled direction and evaluate the CDF
	Float error = evaluate_line_sampling_cdf(line.normal, dot(line.solid.line_direction, sampled_dir), cdf);
	// Compute the backward error
	return error / line.extent_cdf;
}
