#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import itertools


function_decoration = "inline "


def declare_vec(prefix, cpp_scalar_type, dimensionality):
    entries = "xyzwabcd"[0:dimensionality]
    code = ""
    code += "struct %svec%d {\n" % (prefix, dimensionality)
    # Declare the array operator
    code += "\t%s%s& operator[](unsigned int entry_index) {\n" % (function_decoration, cpp_scalar_type)
    code += "\t\treturn *((&x) + entry_index);\n"
    code += "\t}\n"
    code += "\t%s%s operator[](unsigned int entry_index) const {\n" % (function_decoration, cpp_scalar_type)
    code += "\t\treturn *((&x) + entry_index);\n"
    code += "\t}\n\n"
    # Declare a union to support all swizzles that access a true subset of all
    # entries in order
    code += "\tunion {\n"
    code += "\t\tstruct { %s %s; };\n" % (cpp_scalar_type, ", ".join(entries))
    synonym_dict = dict([(entry, 1) for entry in entries])
    for split_count in range(1, dimensionality - 1):
        for splits in itertools.combinations(range(1, dimensionality), split_count):
            splits = [0] + list(splits) + [dimensionality]
            split_code = "\t\tstruct { "
            new_swizzle = False
            for j in range(split_count + 1):
                split_entries = entries[splits[j]:splits[j + 1]]
                suffix_index = synonym_dict.get(split_entries, 0)
                suffix = "_%d" % suffix_index if suffix_index > 0 else ""
                new_swizzle |= (suffix_index == 0)
                synonym_dict[split_entries] = suffix_index + 1
                split_type = "%svec%d" % (prefix, len(split_entries)) if len(split_entries) > 1 else cpp_scalar_type
                split_code += "%s %s%s; " % (split_type, split_entries, suffix)
            split_code += "};\n"
            if new_swizzle:
                code += split_code
    code += "\t};\n"
    code += "};\n\n"
    return code


def declare_vec_constructors(prefix, cpp_scalar_type, dimensionality):
    """
    Declares constructor-like functions for the specified vector type. The
    functions are prefixed with make. Actual constructors are not an option
    because they clash with unions.
    """
    code = ""
    entries = "xyzwabcd"[0:dimensionality]
    # The signature without parameter list
    return_type = "%svec%d" % (prefix, dimensionality)
    signature = "%s%s make_%s" % (function_decoration, return_type, return_type)
    # Declare a broadcasting constructor
    code += "%s(%s scalar) {\n" % (signature, cpp_scalar_type)
    code += "\t%s result;\n" % return_type
    for entry in entries:
        code += "\tresult.%s = scalar;\n" % entry
    code += "\treturn result;\n"
    code += "}\n"
    # For other constructors, which may take vectors, iterate over all ways to
    # write dimensionality as sum of natural numbers
    handled_partitions = set()
    for partition in itertools.product(* dimensionality * [range(dimensionality)]):
        if dimensionality != sum(partition):
            continue
        partition = tuple([size for size in partition if size > 0])
        if partition in handled_partitions:
            continue
        handled_partitions.add(partition)
        remaining_entries = entries
        parts = list()
        for part in partition:
            parts.append(remaining_entries[:part])
            remaining_entries = remaining_entries[part:]
        parameters = [(cpp_scalar_type if len(part) == 1 else "%svec%d" % (prefix, len(part))) + " %s" % part for part in parts]
        code += "%s(%s) {\n" % (signature, ", ".join(parameters))
        code += "\t%s result;\n" % return_type
        for part in parts:
            code += "\tresult.%s = %s;\n" % (part, part)
        code += "\treturn result;\n"
        code += "}\n"
    code += "\n"
    return code


def declare_vec_unary_operator(prefix, cpp_scalar_type, dimensionality, unary_operator):
    vec_type = "%svec%d" % (prefix, dimensionality)
    entries = "xyzwabcd"[0:dimensionality]
    code = ""
    code += "%s%s operator%s(const %s& rhs) {\n" % (function_decoration, vec_type, unary_operator, vec_type)
    code += "\t%s result;\n" % vec_type
    for entry in entries:
        code += "\tresult.%s = %srhs.%s;\n" % (entry, unary_operator, entry)
    code += "\treturn result;\n"
    code += "}\n\n"
    return code


def declare_vec_binary_operator(prefix, cpp_scalar_type, dimensionality, binary_operator):
    vec_type = "%svec%d" % (prefix, dimensionality)
    entries = "xyzwabcd"[0:dimensionality]
    code = ""
    # Scalar as LHS
    code += "%s%s operator%s(%s lhs, const %s& rhs) {\n" % (function_decoration, vec_type, binary_operator, cpp_scalar_type, vec_type)
    code += "\t%s result;\n" % vec_type
    for entry in entries:
        code += "\tresult.%s = lhs %s rhs.%s;\n" % (entry, binary_operator, entry)
    code += "\treturn result;\n"
    code += "}\n\n"
    # Scalar as RHS
    code += "%s%s operator%s(const %s& lhs, %s rhs) {\n" % (function_decoration, vec_type, binary_operator, vec_type, cpp_scalar_type)
    code += "\t%s result;\n" % vec_type
    for entry in entries:
        code += "\tresult.%s = lhs.%s %s rhs;\n" % (entry, entry, binary_operator)
    code += "\treturn result;\n"
    code += "}\n\n"
    # Vectors on both sides
    code += "%s%s operator%s(const %s& lhs, const %s& rhs) {\n" % (function_decoration, vec_type, binary_operator, vec_type, vec_type)
    code += "\t%s result;\n" % vec_type
    for entry in entries:
        code += "\tresult.%s = lhs.%s %s rhs.%s;\n" % (entry, entry, binary_operator, entry)
    code += "\treturn result;\n"
    code += "}\n\n"
    # Scalar as RHS in assignment
    code += "%sconst %s& operator%s=(%s& lhs, %s rhs) {\n" % (function_decoration, vec_type, binary_operator, vec_type, cpp_scalar_type)
    for entry in entries:
        code += "\tlhs.%s %s= rhs;\n" % (entry, binary_operator)
    code += "\treturn lhs;\n"
    code += "}\n\n"
    # Vector as RHS in assignment
    code += "%sconst %s& operator%s=(%s& lhs, const %s& rhs) {\n" % (function_decoration, vec_type, binary_operator, vec_type, vec_type)
    for entry in entries:
        code += "\tlhs.%s %s= rhs.%s;\n" % (entry, binary_operator, entry)
    code += "\treturn lhs;\n"
    code += "}\n\n"
    return code


def declare_mat(prefix, cpp_scalar_type, rows, cols):
    code = ""
    mat_type = "%smat%d" % (prefix, rows) if rows == cols else "%smat%dx%d" % (prefix, rows, cols)
    col_type = "%svec%d" % (prefix, rows)
    code += "struct %s {\n" % mat_type
    indices = [(j, k) for k in range(cols) for j in range(rows)]
    # Declare a default constructor that does nothing
    code += "\t%s%s() {}\n" % (function_decoration, mat_type)
    # Declare a broadcast constructor for the diagonal
    code += "\t%s%s(%s diagonal) {\n" % (function_decoration, mat_type, cpp_scalar_type)
    for j, k in indices:
        code += "\t\tcolumns[%d][%d] = %s;\n" % (k, j, "0" if j != k else "diagonal")
    code += "\t}\n"
    # Declare a constructor that takes all entries individually
    parameters = ["%s col_%d_row_%d" % (cpp_scalar_type, k, j) for j, k in indices]
    code += "\t%s%s(%s) {\n" % (function_decoration, mat_type, ", ".join(parameters))
    for j, k in indices:
        code += "\t\tcolumns[%d][%d] = col_%d_row_%d;\n" % (k, j, k, j)
    code += "\t}\n"
    # Declare a constructor that takes column vectors
    parameters = ["%s col_%d" % (col_type, k) for k in range(cols)]
    code += "\t%s%s(%s) {\n" % (function_decoration, mat_type, ", ".join(parameters))
    for k in range(cols):
        code += "\t\tcolumns[%d] = col_%d;\n" % (k, k)
    code += "\t}\n"
    # Declare the array operator
    code += "\t%s%s& operator[](unsigned int col) {\n" % (function_decoration, col_type)
    code += "\t\treturn columns[col];\n"
    code += "\t}\n"
    code += "\t%sconst %s& operator[](unsigned int col) const{\n" % (function_decoration, col_type)
    code += "\t\treturn columns[col];\n"
    code += "\t}\n\n"
    # Columns are held as vectors
    code += "\t%s columns[%d];\n" % (col_type, cols)
    code += "};\n\n"
    return code


def declare_mat_unary_operator(prefix, cpp_scalar_type, rows, cols, unary_operator):
    mat_type = "%smat%d" % (prefix, rows) if rows == cols else "%smat%dx%d" % (prefix, rows, cols)
    code = ""
    code += "%s%s operator%s(const %s& rhs) {\n" % (function_decoration, mat_type, unary_operator, mat_type)
    code += "\t%s result;\n" % mat_type
    for k in range(cols):
        code += "\tresult.columns[%d] = %srhs[%d];\n" % (k, unary_operator, k)
    code += "\treturn result;\n"
    code += "}\n\n"
    return code


def declare_mat_binary_operator(prefix, cpp_scalar_type, rows, cols, binary_operator):
    mat_type = "%smat%d" % (prefix, rows) if rows == cols else "%smat%dx%d" % (prefix, rows, cols)
    code = ""
    # Scalar as LHS
    code += "%s%s operator%s(%s lhs, const %s& rhs) {\n" % (function_decoration, mat_type, binary_operator, cpp_scalar_type, mat_type)
    code += "\t%s result;\n" % mat_type
    for k in range(cols):
        code += "\tresult.columns[%d] = lhs %s rhs.columns[%d];\n" % (k, binary_operator, k)
    code += "\treturn result;\n"
    code += "}\n\n"
    # Scalar as RHS
    code += "%s%s operator%s(const %s& lhs, %s rhs) {\n" % (function_decoration, mat_type, binary_operator, mat_type, cpp_scalar_type)
    code += "\t%s result;\n" % mat_type
    for k in range(cols):
        code += "\tresult.columns[%d] = lhs.columns[%s] %s rhs;\n" % (k, k, binary_operator)
    code += "\treturn result;\n"
    code += "}\n\n"
    # Vectors on both sides
    code += "%s%s operator%s(const %s& lhs, const %s& rhs) {\n" % (function_decoration, mat_type, binary_operator, mat_type, mat_type)
    code += "\t%s result;\n" % mat_type
    for k in range(cols):
        code += "\tresult.columns[%d] = lhs.columns[%d] %s rhs.columns[%d];\n" % (k, k, binary_operator, k)
    code += "\treturn result;\n"
    code += "}\n\n"
    # Scalar as RHS in assignment
    code += "%sconst %s& operator%s=(%s& lhs, %s rhs) {\n" % (function_decoration, mat_type, binary_operator, mat_type, cpp_scalar_type)
    for k in range(cols):
        code += "\tlhs.columns[%d] %s= rhs;\n" % (k, binary_operator)
    code += "\treturn lhs;\n"
    code += "}\n\n"
    # Vector as RHS in assignment
    code += "%sconst %s& operator%s=(%s& lhs, const %s& rhs) {\n" % (function_decoration, mat_type, binary_operator, mat_type, mat_type)
    for k in range(cols):
        code += "\tlhs.columns[%d] %s= rhs.columns[%d];\n" % (k, binary_operator, k)
    code += "\treturn lhs;\n"
    code += "}\n\n"
    return code


def declare_mat_vec_product(prefix, cpp_scalar_type, rows, cols):
    """
    Declares a matrix-vector multiplication operator where the LHS is a matrix
    of the given size.
    """
    mat_type = "%smat%d" % (prefix, rows) if rows == cols else "%smat%dx%d" % (prefix, rows, cols)
    col_type = "%svec%d" % (prefix, rows)
    row_type = "%svec%d" % (prefix, cols)
    code = ""
    code += "%s%s operator*(const %s& lhs, const %s& rhs) {\n" % (function_decoration, col_type, mat_type, row_type)
    code += "\t%s result;\n" % col_type
    for j in range(rows):
        terms = ["lhs[%d][%d] * rhs[%d]" % (k, j, k) for k in range(cols)]
        code += "\tresult[%d] = %s;\n" % (j, " + ".join(terms))
    code += "\treturn result;\n"
    code += "}\n\n"
    return code


def declare_mat_transpose(prefix, rows, cols):
    mat_type = "%smat%d" % (prefix, rows) if rows == cols else "%smat%dx%d" % (prefix, rows, cols)
    out_type = "%smat%d" % (prefix, rows) if rows == cols else "%smat%dx%d" % (prefix, cols, rows)
    code = ""
    code += "%s%s transpose(const %s& rhs) {\n" % (function_decoration, out_type, mat_type)
    code += "\t%s result;\n" % mat_type
    for k in range(cols):
        for j in range(rows):
            code += "\tresult.columns[%d][%d] = rhs.columns[%d][%d];\n" % (j, k, k, j)
    code += "\treturn result;\n"
    code += "}\n\n"
    return code


def generate_mini_glm():
    """
    :return: C++ code for a header that implements some of the types and methods
        built into GLSL.
    """
    code = "// This file is automatically generated. It implements GLM-like functionality\n"
    code += "// but is less complete. The benefit is that it is a painless dependency.\n"
    code += "#pragma once\n\n"
    # code += "#include <cstdint>\n"
    # types = [("u", "std::uint32_t"), ("", "float")]
    types = [("", "Float")]
    dimensionalities = [2, 3, 4]
    matrix_sizes = [(2, 2)]
    unary_operators = ["-"]
    binary_operators = ["+", "-", "*", "/"]
    for prefix, cpp_scalar_type in types:
        for dimensionality in dimensionalities:
            code += declare_vec(prefix, cpp_scalar_type, dimensionality)
        for rows, cols in matrix_sizes:
            code += declare_mat(prefix, cpp_scalar_type, rows, cols)
    for prefix, cpp_scalar_type in types:
        for dimensionality in dimensionalities:
            code += declare_vec_constructors(prefix, cpp_scalar_type, dimensionality)
            for unary_operator in unary_operators:
                if unary_operator == "-" and prefix == "u":
                    continue
                code += declare_vec_unary_operator(prefix, cpp_scalar_type, dimensionality, unary_operator)
            for binary_operator in binary_operators:
                code += declare_vec_binary_operator(prefix, cpp_scalar_type, dimensionality, binary_operator)
        for rows, cols in matrix_sizes:
            for unary_operator in unary_operators:
                if unary_operator == "-" and prefix == "u":
                    continue
                code += declare_mat_unary_operator(prefix, cpp_scalar_type, rows, cols, unary_operator)
            for binary_operator in binary_operators:
                code += declare_mat_binary_operator(prefix, cpp_scalar_type, rows, cols, binary_operator)
            code += declare_mat_vec_product(prefix, cpp_scalar_type, rows, cols)
            code += declare_mat_transpose(prefix, rows, cols)
    return code


if __name__ == "__main__":
    with open("find_worst_case/mini_glm.h", "w") as file:
        file.write(generate_mini_glm())
